import { z } from 'zod';
import { fetch_with_ratelimit } from './fetchutils';

export const searchProjectsHit = z.object({
	slug: z.string().regex(/^[\w!@$()`.+,"\-']{3,64}$/),
	title: z.string(),
	description: z.string(),
	categories: z.string().array().default([]),
	client_side: z.enum(['required', 'optional', 'unsupported', 'unknown']),
	server_side: z.enum(['required', 'optional', 'unsupported', 'unknown']),
	project_type: z.enum(['mod', 'modpack', 'resourcepack', 'shader']),
	downloads: z.number().int(),
	icon_url: z.union([z.string().url(), z.literal('').transform(() => '/favicon.png')]),
	color: z.number().int().nullable(),
	project_id: z.string(),
	author: z.string(),
	display_categories: z.string().array().default([]),
	versions: z.string().array(),
	follows: z.number().int(),
	date_created: z.string().datetime({ offset: true }),
	date_modified: z.string().datetime({ offset: true }),
	latest_version: z.string().optional(),
	license: z.string(),
	gallery: z.string().array().default([]),
	featured_gallery: z.string().nullable()
});

export const searchProjectsResponse = z.object({
	hits: searchProjectsHit.array(),
	offset: z.number().int(),
	limit: z.number().int(),
	total_hits: z.number().int()
});

const searchProjectsRequest = z.object({
	query: z.string().optional(),
	facets: z
		.string()
		.regex(/^(categories|versions|license|project_type):(.+)$/)
		.array()
		.array()
		.default([]),
	index: z.enum(['relevance', 'downloads', 'follows', 'newest', 'updated']).default('relevance'),
	offset: z.number().int().default(0),
	limit: z.number().int().default(10),
	filters: z.string().optional()
});

function searchProjectsRequestToUrlParams(
	request: z.infer<typeof searchProjectsRequest>
): URLSearchParams {
	const params = new URLSearchParams();
	if (typeof request.query != 'undefined') {
		params.set('query', request.query);
	}
	params.set('facets', JSON.stringify(request.facets));
	params.set('index', request.index);
	params.set('offset', request.offset.toString());
	params.set('limit', request.limit.toString());
	if (typeof request.filters != 'undefined') {
		params.set('filters', request.filters);
	}
	return params;
}

export async function searchProjects(
	request: z.input<typeof searchProjectsRequest>
): Promise<z.infer<typeof searchProjectsResponse>> {
	const parsedRequest = searchProjectsRequest.parse(request);
	const params = searchProjectsRequestToUrlParams(parsedRequest);
	const response = await fetch_with_ratelimit(
		'modrinth',
		'https://api.modrinth.com/v2/search?' + params.toString()
	);
	return searchProjectsResponse.parse(await response.json());
}

export const projectVersion = z.object({
	name: z.string(),
	version_number: z.string(),
	changelog: z.string().nullable(),
	dependencies: z
		.object({
			version_id: z.string().nullable(),
			project_id: z.string().nullable(),
			file_name: z.string().nullable(),
			dependency_type: z.enum(['required', 'optional', 'incompatible', 'embedded'])
		})
		.array(),
	game_versions: z.string().array(),
	version_type: z.enum(['release', 'beta', 'alpha']),
	loaders: z.string().array(),
	featured: z.boolean(),
	status: z.enum(['listed', 'archived', 'draft', 'unlisted', 'scheduled', 'unknown']),
	requested_status: z.enum(['listed', 'archived', 'draft', 'unlisted']).nullable(),
	id: z.string(),
	project_id: z.string(),
	author_id: z.string(),
	date_published: z.string().datetime({ offset: true }),
	downloads: z.number().int(),
	changelog_url: z.null(),
	files: z
		.object({
			hashes: z.object({
				sha512: z.string().regex(/^[a-f0-9]{128}$/),
				sha1: z.string().regex(/^[a-f0-9]{40}$/)
			}),
			url: z.string().url(),
			filename: z.string(),
			primary: z.boolean(),
			size: z.number().int(),
			file_type: z.enum(['required-resource-pack', 'optional-resource-pack']).nullable()
		})
		.array()
});

export const listProjectVersionsResponse = projectVersion.array();

export async function listProjectVersions(
	id: string
): Promise<z.infer<typeof listProjectVersionsResponse>> {
	const response = await fetch_with_ratelimit(
		'modrinth',
		`https://api.modrinth.com/v2/project/${id}/version`
	);
	return listProjectVersionsResponse.parse(await response.json());
}
